"""Sets up a flask-restx server for running the swabseq analysis R-script."""

import os

from flask import Flask
from flask.json import JSONEncoder
from flask_cors import CORS
from flask_restx import Api
# https://github.com/noirbizarre/flask-restplus/issues/565#issuecomment-562610603
from werkzeug.middleware.proxy_fix import ProxyFix


app = Flask(__name__, static_folder='static', static_url_path='')
app.wsgi_app = ProxyFix(app.wsgi_app)


# Configuration ---------------------------------------------------------------

app.config['AUTH_PROVIDER'] = os.environ.get('AUTH_PROVIDER', 'auth0')
app.config['AUTH0_DOMAIN'] = os.environ.get('AUTH0_DOMAIN', '')
app.config['AUTH0_CLIENT_ID'] = os.environ.get('AUTH0_CLIENT_ID', 'Msk8I4Ad2ujE76MwOatsmmvEEds5v50h')
app.config['AUTH0_API_AUDIENCE'] = os.environ.get('AUTH0_API_AUDIENCE', '')
app.config['AUTH0_AUTHORIZATION_URL'] = os.environ.get('AUTH0_AUTHORIZATION_URL', '')
app.config['AUTH0_TOKEN_URL'] = os.environ.get('AUTH0_TOKEN_URL', '')
app.config['RESTX_JSON'] = {'cls': JSONEncoder}  # Add support for serializing datetime/date
app.config['SERVER_VERSION'] = os.environ.get('SERVER_VERSION', 'local')
app.config['SERVER_NAME'] = os.environ.get('SERVER_NAME', app.config.get('SERVER_NAME'))
# app.config['RSCRIPT_THREADS'] = int(os.environ.get('RSCRIPT_THREADS', '8'))
# app.config['CELERY_RESULT_BACKEND'] = os.environ.get('CELERY_RESULT_BACKEND', None)
# app.config['CELERY_BROKER_URL'] = os.environ.get('CELERY_BROKER_URL', None)


# API Documentation -----------------------------------------------------------

authorizations = {
    'token': {
        'type': 'oauth2',
        'flow': 'accessCode',
        'audience': app.config['AUTH0_API_AUDIENCE'],
        'domain': app.config['AUTH0_DOMAIN'],
        'clientId': app.config['AUTH0_CLIENT_ID'],
        'authorizationUrl': app.config['AUTH0_AUTHORIZATION_URL'],
        'tokenUrl': app.config['AUTH0_TOKEN_URL'],
        'scopes': {}
    }
}
api = Api(
    app,
    title='Swabseq Analysis API',
    version='0.1.0',
    authorizations=authorizations,
)
CORS(app)



#!/usr/bin/env python

"""Sets up a flask-restx server for running the swabseq analysis R-script."""

from script_runner.server import app, api
from script_runner.api.script import api as script
from script_runner.api.health import api as server_health


api.add_namespace(script)
api.add_namespace(server_health)


if __name__ == '__main__':
    app.run()



import base64
from script_runner.config import settings
import csv
import glob
import json
import re
import tempfile
import traceback
import os
import shutil
import subprocess
from celery import Celery
from distutils.dir_util import copy_tree
import jsonpath_ng


# Celery ----------------------------------------------------------------------

celery = Celery(
    'swabseq-analysis-celery',
    backend=os.environ.get('CELERY_RESULT_BACKEND', None),
    broker=os.environ.get('CELERY_BROKER_URL', None),
)

# celery.conf.update(app.config)


# Analysis --------------------------------------------------------------------

def result_is_valid(result, skip_lists):
    for name, values in skip_lists.items():
        col_value = result.get(name, None)
        if col_value is None or col_value in values:
            return False
    return True

def b64encode_file(filepath):
    with open(filepath, "rb") as input_file:
        return base64.b64encode(input_file.read()).decode('utf-8')

def read_csv_as_dict_list(filepath, filter_fn=None):
    with open(filepath, "r") as csv_file:
        csv_reader = csv.DictReader(csv_file)
        if filter_fn is None:
            filter_fn = lambda x: True
        return [x for x in csv_reader if filter_fn(x)]

def read_json_file(filepath):
    with open(filepath, "r") as json_file:
        return json.load(json_file)

def rename_fields(original, fields):
    return {
        new_field: original[original_field]
        for new_field, original_field
        in fields.items()
        if original_field in original
    }

def replace_env_vars(input, env_vars):
    for name, value in env_vars.items():
        input = input.replace(f"${name}", value)
        # input = input.replace(f"${{{name}}}", value)
    return input

def build_command(rundir, request_path, request_query_params, request_body):
    env_vars = {'RUNDIR': rundir}
    for path_arg in settings.inputs.path_args:
        path_match = re.match(path_arg.path_regex, request_path)
        path_match_group = path_arg.path_regex_match_group if path_arg.path_regex_match_group else 0
        if path_match:
            env_vars[path_arg.env_var_name] = path_match.group(path_match_group)
    for query_arg in settings.inputs.query_args:
        query_arg_values = request_query_params.get(query_arg.query_var_name, "")
        if len(query_arg_values) == 1:
            env_vars[query_arg.env_var_name] = query_arg_values[0]
        elif len(query_arg_values) > 1:
            env_vars[query_arg.env_var_name] = query_arg_values
    for body_arg in settings.inputs.body_args:
        jsonpath_expression = jsonpath_ng.parse(body_arg.body_jsonpath)
        env_vars[body_arg.env_var_name] = jsonpath_expression.find(request_body)
    return [
        replace_env_vars(command_arg, env_vars)
        for command_arg
        in settings.command
    ]

def build_response(rundir, status='ready'):
    attachments = {}
    for attachment_glob in settings.outputs.attachments:
        for attachment in glob.glob(f"{rundir}/{attachment_glob}"):
            attachments[os.path.basename(attachment)] = b64encode_file(attachment)

    response = {
        'status': status,
        'attachments': attachments,
    }

    # TODO: Add input args to response as keys.

    if settings.outputs.results.format == 'csv':
        response['results'] = read_csv_as_dict_list(os.path.join(rundir, settings.outputs.results.input_path))
        if settings.outputs.results.remapped_columns is not None:
            remapped_columns = {}
            skip_lists = {}
            for remapped_column in settings.outputs.results.remapped_columns:
                remapped_columns[remapped_column.output_name] = remapped_column.input_name
                skip_lists[remapped_column.output_name] = set(remapped_column.skip_list) if remapped_column.skip_list is not None else set()
            results = []
            for row in response['results']:
                result = rename_fields(row, remapped_columns)
                if result_is_valid(result, skip_lists):
                    results.append(result)
            response['results'] = results
    elif settings.outputs.results.format == 'json':
        response['results'] = read_json_file(os.path.join(rundir, settings.outputs.results.input_path))

    return response

def do_script(rundir, request_path, request_query_params, request_body):
    os.makedirs(os.path.join(rundir, "out"))

    script_args = build_command(rundir, request_path, request_query_params, request_body)
    subprocess.check_call(script_args)
    return build_response(rundir, status='ready')

def log_disk_usage(path = None):
    if path is None:
        path = os.getcwd()
    stat = shutil.disk_usage(path)
    print(f"Disk usage statistics for '{path}':")
    print(stat)

def copy_command_rundir_base(rundir):
    if settings.command_rundir_base:
        print(f"Copying base files from {settings.command_rundir_base}")
        try:
            copy_tree(settings.command_rundir_base, rundir)
        except Exception as ex:
            print(f"Failed to copy command_base_dir ({settings.command_rundir_base}) to {rundir}")
            print(traceback.format_exc())

@celery.task()
def run_script(request_path, request_query_params, request_body):
    try:
        command_rundir = settings.command_rundir if settings.command_rundir is not None else os.getcwd()

        # Run R script and zip results to generate temp file
        if settings.debug:
            # If we're in debug mode, don't delete the work directory
            rundir = tempfile.TemporaryDirectory(prefix="results-", dir=command_rundir).name
            copy_command_rundir_base(rundir)
            log_disk_usage()
            try:
                return do_script(rundir, request_path, request_query_params, request_body)
            finally:
                log_disk_usage()
        else:
            with tempfile.TemporaryDirectory(prefix="results-", dir=command_rundir) as rundir:
                copy_command_rundir_base(rundir)
                log_disk_usage()
                try:
                    return do_script(rundir, request_path, request_query_params, request_body)
                finally:
                    log_disk_usage()
    except Exception as ex:
        ex_str = traceback.format_exc()
        print(ex_str)
        return {
            'status': 'failed',
            'error': ex_str,
        }



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal
from pydantic import BaseSettings, BaseModel

def json_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a JSON file
    at the project's root.

    Here we happen to choose to use the `env_file_encoding` from Config
    when reading `config.json`
    """
    encoding = settings.__config__.env_file_encoding
    return json.loads(Path('config.json').read_text(encoding))

class InputPathArg(BaseModel):
    env_var_name: str
    path_regex: str
    path_regex_match_group: int

class InputQueryArg(BaseModel):
    env_var_name: str
    query_var_name: str

class InputBodyArg(BaseModel):
    env_var_name: str
    body_jsonpath: str

class InputSettings(BaseModel):
    path_args: List[InputPathArg]
    query_args: List[InputQueryArg]
    body_args: List[InputBodyArg]

class JSONOutput(BaseModel):
    kind: Literal['stdout', 'stderr', 'file']
    format: Literal['json']

    input_path: str

class CSVOutputColumn(BaseModel):
    output_name: str
    input_name: str

    skip_list: Optional[List[str]]

class CSVOutput(BaseModel):
    kind: Literal['stdout', 'stderr', 'file']
    format: Literal['csv']

    input_path: str
    remapped_columns: Optional[List[CSVOutputColumn]]

class OutputSettings(BaseModel):
    results: Union[JSONOutput, CSVOutput]
    attachments: List[str]

class Settings(BaseSettings):
    celery_result_backend: Optional[str]
    celery_broker_url: Optional[str]

    command: List[str]
    command_rundir: Optional[str]
    command_rundir_base: Optional[str]
    inputs: InputSettings
    outputs: OutputSettings

    debug: Optional[bool]

    class Config:
        env_file_encoding = 'utf-8'

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                json_config_settings_source,
                env_settings,
                file_secret_settings,
            )


settings = Settings()



from flask_restx import Resource, fields, Namespace

from script_runner.server import app


api = Namespace('server-health', description="Service health checks.", path='/')


health_output = api.model('HealthOutput', {
    'version': fields.String(),
    'server': fields.Boolean(),
})


@api.route('/health')
class HealthResource(Resource):
    @api.doc(model=health_output)
    def get(self):
        return {
            'version': app.config['SERVER_VERSION'],
            'server': True,
        }



from celery.result import AsyncResult

from flask import abort, request
from flask_restx import Resource, fields, Namespace

from script_runner.analysis import run_script
from script_runner.authorization import requires_auth


api = Namespace('script', description='Operations for running a long-lived script.', path='/')


script_result = api.model('ScriptResult', {})
script_attachments = api.model('ScriptAttachments', {
    'LIMS_results.csv': fields.String(),
})
script_output = api.model('ScriptOutput', {
    'id': fields.String(),
    'status': fields.String(),
    'results': fields.List(fields.Nested(script_result)),
    'attachments': fields.Nested(script_attachments),
})


@api.route('/script/<path:path>')
class ScriptTasksResource(Resource):
    @api.doc(security='token', model=script_output)
    @requires_auth
    def post(self, path):
        task = run_script.delay(path, request.args, request.json)

        return {
            "id": task.task_id,
            "status": "ready" if task.ready() else "not-ready",
        }


@api.route('/script/<string:task_id>')
class ScriptTaskResource(Resource):
    @api.doc(security='token', model=script_output)
    @requires_auth
    def get(self, task_id):
        result = AsyncResult(task_id)

        if result.ready():
            response = result.get()
            response['id'] = result.task_id
            return response
        else:
            return {
                'id': result.task_id,
                'status': 'not-ready',
            }



"""Utilities for API authorization."""

import json

from functools import wraps
from six.moves.urllib.request import urlopen
from jose import jwt

from flask import request
from flask_cors import cross_origin

from script_runner.server import app, api


# Auth Errors -----------------------------------------------------------------

class AuthError(Exception):
    """Represents a failure during authentication"""
    def __init__(self, error, status_code=401):
        super(AuthError, self).__init__(error)
        self.error = error
        self.status_code = status_code

@cross_origin()
@api.errorhandler(AuthError)
def handle_auth_error(ex):
    """Flask error handler"""
    return ex.error, ex.status_code


# Auth Helpers ----------------------------------------------------------------

def get_token_auth_header():
    """Obtains the Access Token from the Authorization Header"""
    auth = request.headers.get("Authorization", None)
    if not auth:
        raise AuthError({
            "code": "authorization_header_missing",
            "description": "Authorization header is expected"
        }, 401)

    parts = auth.split()

    if parts[0].lower() != "bearer":
        raise AuthError({
            "code": "invalid_header",
            "description": "Authorization header must start with Bearer"
        }, 401)
    if len(parts) == 1:
        raise AuthError({
            "code": "invalid_header",
            "description": "Token not found"
        }, 401)
    if len(parts) > 2:
        raise AuthError({
            "code": "invalid_header",
            "description": "Authorization header must be Bearer token"
        }, 401)

    token = parts[1]
    return token

def decode_jwt():
    token = get_token_auth_header()
    jsonurl = urlopen("https://"+app.config['AUTH0_DOMAIN']+"/.well-known/jwks.json")
    jwks = json.loads(jsonurl.read())
    unverified_header = jwt.get_unverified_header(token)
    rsa_key = {}
    for key in jwks["keys"]:
        if key["kid"] == unverified_header["kid"]:
            rsa_key = {
                "kty": key["kty"],
                "kid": key["kid"],
                "use": key["use"],
                "n": key["n"],
                "e": key["e"]
            }
    if rsa_key:
        try:
            return jwt.decode(
                token,
                rsa_key,
                algorithms=["RS256"],
                audience=app.config['AUTH0_API_AUDIENCE'],
                issuer="https://"+app.config['AUTH0_DOMAIN']+"/"
            )
        except jwt.ExpiredSignatureError:
            raise AuthError({
                "code": "token_expired",
                "description": "token is expired"
            }, 401)
        except jwt.JWTClaimsError:
            raise AuthError({
                "code": "invalid_claims",
                "description": "incorrect claims, please check the audience and issuer"
            }, 401)
        except Exception:
            raise AuthError({
                "code": "invalid_header",
                "description": "Unable to parse authentication token."
            }, 401)


# Decorators ------------------------------------------------------------------

def requires_auth(f):
    """Determines if the Access Token is valid"""
    @wraps(f)
    def decorated(*args, **kwargs):
        if app.config['AUTH_PROVIDER'] == 'none':
            request.current_user = {'sub': '42'}
            return f(*args, **kwargs)
        elif app.config['AUTH_PROVIDER'] == 'auth0':
            payload = decode_jwt()
            if payload:
                request.current_user = payload
                return f(*args, **kwargs)
            raise AuthError({
                "code": "invalid_header",
                "description": "Unable to find appropriate key"
            }, 401)
        return f(*args, **kwargs)
    return decorated



import os
from distutils.core import setup

def get_version():
    """
    Gets the latest version number out of the package,
    saving us from maintaining it in multiple places.
    """
    local_results = {}
    with open('script_runner/version.py') as version_file:
        exec(version_file.read(), {}, local_results)
    return local_results['__version__']

setup(
    name="script-runner-api",
    packages=[
        "script_runner",
        "script_runner.api",
    ],
    scripts=["script_runner/main.py"],
    version=get_version(),
    license="MIT",
    description="Provides an API server + celery worker module for running arbitrary scripts.",
    author="Robert Chu",
    author_email="robert.chu@labgrid.com",
    url="https://github.com/lab-grid/script-runner",
    download_url="https://github.com/lab-grid/script-runner/archive/0.1.0.tar.gz",
    keywords=["script", "api", "labflow", "labgrid", "lab-grid"],
    install_requires=[
        "amqp==5.0.6; python_version >= '3.6'",
        "aniso8601==9.0.1; python_version >= '3.5'",
        "attrs==20.3.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "billiard==3.6.4.0",
        "celery==5.0.5",
        "click==7.1.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "click-didyoumean==0.0.3",
        "click-plugins==1.1.1",
        "click-repl==0.1.6",
        "decorator==5.0.5; python_version >= '3.5'",
        "ecdsa==0.14.1; python_version >= '2.6' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "flask==1.1.2",
        "flask-cors==3.0.10",
        "flask-restx==0.3.0",
        "gunicorn==20.1.0",
        "itsdangerous==1.1.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "jinja2==2.11.3; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "jsonpath-ng==1.5.2",
        "jsonschema==3.2.0",
        "kombu==5.0.2; python_version >= '3.6'",
        "markupsafe==1.1.1; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "ply==3.11",
        "prompt-toolkit==3.0.18; python_full_version >= '3.6.1'",
        "pyasn1==0.4.8",
        "pydantic==1.8.1",
        "pyrsistent==0.17.3; python_version >= '3.5'",
        "python-jose==3.2.0",
        "pytz==2021.1",
        "redis==3.5.3",
        "rsa==4.7.2; python_version >= '3.5' and python_version < '4'",
        "six==1.15.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "typing-extensions==3.7.4.3",
        "vine==5.0.0; python_version >= '3.6'",
        "wcwidth==0.2.5",
        "werkzeug==1.0.1",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
